/**
 * Copyright 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 */
package com.sunlabs.archsim.fleet.ships.fetchAndIssue;


import java.util.LinkedList;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.codeStorage.CodeBag;
import com.sunlabs.archsim.fleet.codeStorage.CodeStorage;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.globals.Globals;

/**
 * @author ib27688
 *
 */
public class FetchAndIssue extends Component {
	
	protected Globals globals= Globals.getInstance();

	protected IssueComponent instructionIssueComponent;
	protected IssueComponent literalIssueComponent;
	protected FetchComponent fetchComponent;
	
	protected Terminal fetchInTerm;
	protected Terminal ackFetchInTerm;
	
	protected Terminal instrOutTerm;
	protected Terminal ackInstrOutTerm;
	
	protected Terminal literalOutTerm;
	protected Terminal ackLiteralOutTerm;
	
	protected CodeStorage codeStorage;
	
	public FetchAndIssue(String name, Parameters params) {
		super(name, null);
		

		// process the parameters
		if (params != null) {
			for (int i= 0; i<params.size(); i++) {
				String pName= params.getName(i);
				String sValue= params.getValue(i);
		
				director.error("Component: " + name + ", id= " + getID() 
						+ ": Attribute " + pName + " with value " + sValue
						+ " is ignored");					
		
			} // for
		} // if params != null

		buildFetchAndIssue();
		
		
	} // FetchIssueShip constructor
	
	protected void buildFetchAndIssue() {
		// make the code storage
		codeStorage= CodeStorage.getInstance();
		
		// TODO assume code storage is populated
		
		// make the components
		// note, this is a place to pass parameters to IssueComponent!!
		instructionIssueComponent= new IssueComponent(name+"InstructionIssueComponent", null);
		addMember(instructionIssueComponent);
		// note, this is a place to pass parameters to IssueComponent!!
		literalIssueComponent= new IssueComponent(name+"LiteralIssueComponent", null);
		addMember(literalIssueComponent);
		
		// TODO note, this is a place to pass parameters to FETCHComponent!!
		fetchComponent= new FetchComponent(name+"FetchComponent",
				instructionIssueComponent,
				literalIssueComponent);
		addMember(fetchComponent);

		
		// set delays
		Integer fDelay= globals.intValue(FetchAndIssueDefaults.INSTRUCTION_ISSUE_DELAY);
		Delay functionDelay= (fDelay != null)
			? new Delay(fDelay.intValue())
			: FetchAndIssueDefaults.INSTRUCTION_ISSUE_DELAY_DEF;   
		// set the delay
		instructionIssueComponent.issueAction.issueFn.setDelay(functionDelay);
			
		fDelay= globals.intValue(FetchAndIssueDefaults.LITERAL_ISSUE_DELAY);
		functionDelay= (fDelay != null)
			? new Delay(fDelay.intValue())
			: FetchAndIssueDefaults.LITERAL_ISSUE_DELAY_DEF;   
		// set the delay
		literalIssueComponent.issueAction.issueFn.setDelay(functionDelay);

		fDelay= globals.intValue(FetchAndIssueDefaults.FETCH_ACK_DELAY);
		functionDelay= (fDelay != null)
			? new Delay(fDelay.intValue())
			: FetchAndIssueDefaults.FETCH_ACK_DELAY_DEF;   
		// set the delay
		fetchComponent.fetchAction.fetchAckFn.setDelay(functionDelay);
		
		fetchInTerm= fetchComponent.in;
		ackFetchInTerm= fetchComponent.ackIn;
		instrOutTerm= instructionIssueComponent.out;
		ackInstrOutTerm= instructionIssueComponent.ackOut;
		literalOutTerm= literalIssueComponent.out;
		ackLiteralOutTerm= literalIssueComponent.ackOut;
		
		// rename terminals
		instrOutTerm.setAlias(FetchAndIssueDefaults.INSTRUCTION_OUT);
		ackInstrOutTerm.setAlias(FetchAndIssueDefaults.ACK_INSTRUCTION_OUT);
		literalOutTerm.setAlias(FetchAndIssueDefaults.LITERAL_OUT);
		ackLiteralOutTerm.setAlias(FetchAndIssueDefaults.ACK_LITERAL_OUT);
		
		
		// bootstrap, first get ready the lists for instructions and for literals
		LinkedList<Payload> instrList= new LinkedList<Payload>();
		LinkedList<Payload> literalList= new LinkedList<Payload>();
		// go over all bootstrap code bags and sort payloads into
		// instructions and literals
		for (CodeBag bag: codeStorage.bootstrap) {
			for (Payload p: bag.instructions()) {
				// does the instruction repeat?
				if (p.repeats ==1 || p.repeats < 0) {
					// classify the instruction
					if (p.type == Payload.PayloadType.INSTRUCTION) {
						instrList.add(p);
					}
					else if (p.type == Payload.PayloadType.LITERAL) {
						literalList.add(p);
					}
					else {
						// error
						director.error("Code bag "
								+ bag.getName()
								+ " contains a Payload of unspecified type: "
								+ p);
					}
				}
				else if (p.repeats > 1) {
					// need to take care of cloning if data to make sure this always works
					// it will be OK for Integers and Strings, but if Data.value does not
					// implement a deep clone, two payloads may carry the same data object
					for (int i=0; i< p.repeats; i++) {
						if (p.type == Payload.PayloadType.INSTRUCTION) {
							instrList.add((Payload)p.clone());
						}
						else if (p.type == Payload.PayloadType.LITERAL) {
							literalList.add((Payload)p.clone());
						}
						else {
							// error
							director.error("Code bag "
									+ bag.getName()
									+ " contains a Payload of unspecified type: "
									+ p);
						}	
					}
				} // else
			} // for payload
		} // for bag
		
//		System.out.println("InstructionList: " + instrList);
//		System.out.println("InstructionList: " +literalList);
		
		// pass instructions to the instruction issue component
		if (instrList.size() > 0) {
			instructionIssueComponent.addInstructions(instrList);

		}
		// pass literals to literal issue component
		if (literalList.size() > 0) {
			literalIssueComponent.addInstructions(literalList);

		}
				
	}
	
	public Terminal getTerminal(String name) {
		
		Terminal t= null;
		if (name.equals(fetchInTerm.getBaseName()) || name.equals(fetchInTerm.getBaseAlias())
				|| name.equals(fetchInTerm.getName()) || name.equals(fetchInTerm.getAlias())) {
			t= fetchInTerm;
		}
		else if (name.equals(ackFetchInTerm.getBaseName()) || name.equals(ackFetchInTerm.getBaseAlias())
				|| name.equals(ackFetchInTerm.getName()) || name.equals(ackFetchInTerm.getAlias())) {
			t= ackFetchInTerm;
		}
		else if (name.equals(instrOutTerm.getBaseName()) || name.equals(instrOutTerm.getBaseAlias())
				|| name.equals(instrOutTerm.getName()) || name.equals(instrOutTerm.getAlias())) {
			t= instrOutTerm;
		}
		else if (name.equals(ackInstrOutTerm.getBaseName()) || name.equals(ackInstrOutTerm.getBaseAlias())
				|| name.equals(ackInstrOutTerm.getName()) || name.equals(ackInstrOutTerm.getAlias())) {
			t= ackInstrOutTerm;
		}
		else if (name.equals(literalOutTerm.getBaseName()) || name.equals(literalOutTerm.getBaseAlias())
				|| name.equals(literalOutTerm.getName()) || name.equals(literalOutTerm.getAlias())) {
			t= literalOutTerm;
		}
		else if (name.equals(ackLiteralOutTerm.getBaseName()) || name.equals(ackLiteralOutTerm.getBaseAlias())
				|| name.equals(ackLiteralOutTerm.getName()) || name.equals(ackLiteralOutTerm.getAlias())) {
			t= ackLiteralOutTerm;
		}
		return t;
	} // getTerminal
	
	
} // FetchIssueShip
